#include "TermMaxNoImpGenOp.hpp"
#include "beagle/Beagle.hpp"

using namespace Beagle;

/*!
 *  \brief Construct a termination check operator based on a maximum consecutive generations without improvement of the best-so-far indivudual.
 *  \param inName Name of the operator.
 */
TermMaxNoImpGenOp::TermMaxNoImpGenOp(std::string inName) :
  TerminationOp(inName), dFitnessChange(0)
{	}


/*!
 *  \brief Initialize the maximum generation without improvement termination operator.
 *  \param ioSystem System to use to initialize the operator.
 */
void TermMaxNoImpGenOp::initialize(System& ioSystem)
{
  if(ioSystem.getRegister().isRegistered("ec.term.maxnoimpgen")) {
    mMaxGenerationNoImp =
      castHandleT<UInt>(ioSystem.getRegister().getEntry("ec.term.maxnoimpgen"));
  } else {
    mMaxGenerationNoImp = new UInt(50);
    std::string lLongDescrip = "Maximum number of consecutive generations in which no new best-so-far solution is found. ";
    lLongDescrip += "A zero value means that there is no limit.";
    Register::Description lDescription(
      "Max generation without improvement term criterion",
      "UInt",
      "50",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry("ec.term.maxnoimpgen", mMaxGenerationNoImp, lDescription);
  }
}


/*!
 *  \brief Check if the maximum number of consecutive generations without improvement is reached.
 *  \param inDeme Actual deme of the evolution.
 *  \param ioContext Actual evolution context.
 *  \return True if the ending criterion is reached, false if not.
 */
bool TermMaxNoImpGenOp::terminate(const Deme& inDeme, Context& ioContext)
{
/*	procitaj broj jedinki u hall-of-fame
	pronadji najveci indeks generacije medju svim jedinkama u HOF
	usporedi najveci indeks sa trenutnom generacijom
*/
	unsigned int lHOFSize,i,lMaxGen = 0,lGen, lNoImpGen;
	Beagle::FitnessSimple::Handle lFit;
	double dMaxFt, dMinFt, dFitness, dMaxFt2;
	
	if(mMaxGenerationNoImp->getWrappedValue() == 0) return false;
	lNoImpGen = mMaxGenerationNoImp->getWrappedValue();

	// get max & min population fitness
	dMaxFt = (*(ioContext.getVivarium().getStats()))[0].mMax;
	dMinFt = (*(ioContext.getVivarium().getStats()))[0].mMin;
	if(ioContext.getGeneration() == 1)
	{	dFitnessPast = new double[lNoImpGen];
		dFitnessChange = (dMaxFt - dMinFt)/1000;
	}

	if(ioContext.getGeneration() <= mMaxGenerationNoImp->getWrappedValue())
	{	dFitnessPast[ioContext.getGeneration() - 1] = dMaxFt;
	}
	else
	{	dFitnessNow = dMaxFt;
		std::cout << std::endl << dFitnessNow - dFitnessPast[0] << std::endl;
		if( dFitnessNow - dFitnessPast[0] < dFitnessChange)
		{	delete [] dFitnessPast;
			return true;
		}
		for(unsigned int i=0; i<lNoImpGen-1; i++)
			dFitnessPast[i] = dFitnessPast[i+1];
		dFitnessPast[lNoImpGen-1] = dFitnessNow;
	}

	// get HOF size
	if( (lHOFSize = (unsigned int) ioContext.getVivarium().getHallOfFame().size()) == 0 )
		return false;
	// find max generation index of all HOF members
	for(i=0; i<lHOFSize; i++)
	{	lGen = (ioContext.getVivarium().getHallOfFame())[i].mGeneration;
		// get this hof member fitness
		Fitness::Handle hFit = ( (ioContext.getVivarium().getHallOfFame())[i].mIndividual)->getFitness();
		FitnessSimple FitSimp = castObjectT<Beagle::FitnessSimple&>(*hFit);
		dFitness = FitSimp.getValue();

		if(i==0 || dFitness > dMaxFt2)
			dMaxFt2 = dFitness;
	
		if( lGen > lMaxGen )
			lMaxGen = lGen;
	}

	// is termination overdue
	if( (ioContext.getGeneration()-lMaxGen) > mMaxGenerationNoImp->getWrappedValue() )
	{	
	Beagle_LogInfoM(
      ioContext.getSystem().getLogger(),
      "termination", "Beagle::TermMaxNoImpGenOp",
      std::string("Maximum number of generations without improvement (") +
      uint2str(mMaxGenerationNoImp->getWrappedValue()) +
      std::string(") termination criterion overpassed")
    );
    Beagle_LogDetailedM(
      ioContext.getSystem().getLogger(),
      "termination", "Beagle::TermMaxNoImpGenOp",
      std::string("Actual generation number is: ")+uint2str(ioContext.getGeneration())
    );   
    return true;
	}
	// is termination reached
	if( ((ioContext.getGeneration()-lMaxGen) == mMaxGenerationNoImp->getWrappedValue()) &&
		(ioContext.getDemeIndex() == (ioContext.getVivarium().size()-1)) )
	{
    Beagle_LogInfoM(
      ioContext.getSystem().getLogger(),
      "termination", "Beagle::TermMaxNoImpGenOp",
      std::string("Maximum number of generations without improvement (") +
      uint2str(mMaxGenerationNoImp->getWrappedValue()) +
      std::string(") termination criterion reached")
    );    
    return true;
	}
	// otherwise
  Beagle_LogTraceM(
    ioContext.getSystem().getLogger(),
    "termination", "Beagle::TermMaxNoImpGenOp",
    std::string("Maximum number of generations without improvement (") +
    uint2str(mMaxGenerationNoImp->getWrappedValue()) +
    std::string(") termination criterion not reached ")  
  );
  return false;

}