/*!
 *  \file   beagle/GP/src/WeightedPrimitiveSet.cpp
 *  \brief  Implementation of the type GP::WeightedPrimitiveSet.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/07/01 23:59:00 $
 */

#include "beagle/GP.hpp"
#include "WeightedPrimitiveSet.h"

#define ID_ALL      (UINT_MAX)
#define ID_BRANCHES (UINT_MAX-1)

using namespace Beagle;

/*!
 *  \brief Construct a new primitive set.
 *  \param inRootType Associated tree's root type. Used only with strongly-typed GP.
 */
GP::WeightedPrimitiveSet::WeightedPrimitiveSet(std::string inRootType) :
  mRootType(inRootType)
{ }


/*!
 *  \brief Computer roulette weight of given number of arguments.
 *  \param outRoulette Roulette weight to compute.
 *  \param inNbArgs Number of arguments of primitives associated to the roulette.
 */
void GP::WeightedPrimitiveSet::computeRoulette(GP::WeightedPrimitiveSet::Roulette& outRoulette,
                                       unsigned int inNbArgs) const
{
  outRoulette.resize(0);
  double lSumWeigths = 0.0;
  if(inNbArgs == ID_BRANCHES) {
    for(unsigned int i=0; i<mWeights.size(); ++i) {
      if(mWeights[i].second->getNumberArguments() != 0) {
        outRoulette.push_back(mWeights[i]);
        lSumWeigths += mWeights[i].first;
      }
    }
  }
  else if(inNbArgs == ID_ALL) {
    for(unsigned int i=0; i<mWeights.size(); ++i) {
      outRoulette.push_back(mWeights[i]);
      lSumWeigths += mWeights[i].first;
    }
  }
  else {
    for(unsigned int i=0; i<mWeights.size(); ++i) {
      if(mWeights[i].second->getNumberArguments() == inNbArgs) {
        outRoulette.push_back(mWeights[i]);
        lSumWeigths += mWeights[i].first;
      }
    }
  }

  if(outRoulette.size() == 0) return;
  if(lSumWeigths == 0.0) {
    const double lNormWeigth = 1.0 / outRoulette.size();
    outRoulette[0].first = lNormWeigth;
    for(unsigned int i=1; i<outRoulette.size(); ++i) {
      outRoulette[i].first = outRoulette[i-1].first + lNormWeigth;
    }
  }
  else {
    outRoulette[0].first /= lSumWeigths;
    for(unsigned int i=1; i<outRoulette.size(); ++i) {
      outRoulette[i].first /= lSumWeigths;
      outRoulette[i].first += outRoulette[i-1].first;
    }
  }
}


/*!
 *  \brief Get the associated tree's root type. Used only with strongly-typed GP.
 *  \return Associated tree's root type.
 */
std::string GP::WeightedPrimitiveSet::getRootType() const
{
  return mRootType;
}


/*!
 *  \brief Get the primitive of the name given.
 *  \param inName Name of the primitive to get.
 *  \return Handle to the named primitive, a NULL handle if there is no primitive of the given name.
 */
GP::Primitive::Handle GP::WeightedPrimitiveSet::getPrimitiveByName(std::string inName) const
{
  Map::const_iterator lMapIter = mNames.find(inName);
  if(lMapIter == mNames.end()) return GP::Primitive::Handle(NULL);
  return castHandleT<GP::Primitive>(lMapIter->second);
}


/*!
 *  \brief Initialize the set of primitive.
 *  \param ioSystem System used to initialize the primitive set.
 */
void GP::WeightedPrimitiveSet::initialize(GP::System& ioSystem)
{
  for(unsigned int i=0; i<size(); i++) (*this)[i]->initialize(ioSystem);
}


/*!
 *  \brief Insert a new primitive in the set.
 *  \param inPrimitive Handle to the primitive to insert.
 *  \param inWeight Selection weight associated to primitive.
 *  \throw AssertException If the given handle is NULL.
 *  \throw RunTimeException If a primitive of the same name is already in the set.
 */
void GP::WeightedPrimitiveSet::insert(GP::Primitive::Handle inPrimitive, double inWeight)
{
  Beagle_NonNullPointerAssertM(inPrimitive);
  if(mNames.find(inPrimitive->getName()) != mNames.end()) {
    std::string lMessage = "A primitive named \"";
    lMessage += inPrimitive->getName();
    lMessage += "\" is already in the primitive set.";
    throw Beagle_RunTimeExceptionM(lMessage);
  }
  mNames[inPrimitive->getName()] = inPrimitive;
  (*this).push_back(inPrimitive);
  mWeights.push_back(std::pair<float,GP::Primitive::Handle>(inWeight,inPrimitive));
  mMap[inPrimitive->getNumberArguments()].resize(0);
  mMap[ID_ALL].resize(0);
  if(inPrimitive->getNumberArguments()!=0) mMap[ID_BRANCHES].resize(0);
}


/*!
 *  \brief Select randomly a primitive.
 *  \param ioSystem Reference to the evolution system.
 *  \return Handle to the selected primitive.
 */
GP::Primitive::Handle GP::WeightedPrimitiveSet::selectPrimitive(GP::System& ioSystem)
{
  Roulette& lAll = mMap[ID_ALL];
  if(lAll.size() == 0) computeRoulette(lAll, ID_ALL);
  return selectWithRoulette(lAll, ioSystem);
}


/*!
 *  \brief Select randomly a primitive with arguments, a branch.
 *  \param ioSystem Reference to the evolution system.
 *  \return Handle to the selected primitive, a NULL handle if no branches are in the set.
 */
GP::Primitive::Handle GP::WeightedPrimitiveSet::selectPrimitiveWithArguments(GP::System& ioSystem) {
  Roulette& lBranches = mMap[ID_BRANCHES];
  if(lBranches.size() == 0) computeRoulette(lBranches, ID_BRANCHES);
  return selectWithRoulette(lBranches, ioSystem);
}


/*!
 *  \brief Select randomly a primitive with exactly k arguments.
 *  \param inK Number of arguments of the primitive to select.
 *  \param ioSystem Reference to the evolution system.
 *  \return Handle to the selected primitive, a NULL handle if there is no primitive with k args.
 */
GP::Primitive::Handle
GP::WeightedPrimitiveSet::selectPrimitiveWithKArguments(unsigned int inK, GP::System& ioSystem) {
  Roulette& lRoulette = mMap[inK];
  if(lRoulette.size() == 0) computeRoulette(lRoulette, inK);
  return selectWithRoulette(lRoulette, ioSystem);
}


/*!
 *  \brief Select randomly a primitive without arguments, a leaf.
 *  \param ioSystem Reference to the evolution system.
 *  \return Handle to the selected primitive. a NULL handle if there is no leaf in the set.
 */
GP::Primitive::Handle GP::WeightedPrimitiveSet::selectPrimitiveWithoutArgument(GP::System& ioSystem)
{
  Roulette& lTerminals = mMap[0];
  if(lTerminals.size() == 0) computeRoulette(lTerminals, 0);
  return selectWithRoulette(lTerminals, ioSystem);
}


/*!
 *  \brief Select a primitive from roulette.
 *  \param inRoulette Roulette to select primitive from.
 *  \param ioSystem Evolutionary system.
 */
GP::Primitive::Handle
GP::WeightedPrimitiveSet::selectWithRoulette(const GP::WeightedPrimitiveSet::Roulette& inRoulette,
                                     GP::System& ioSystem) const
{
  if(inRoulette.empty()) return GP::Primitive::Handle(NULL);
  double lVal = ioSystem.getRandomizer().rollUniform(0.0, 1.0);
  unsigned int lSel=0;
  for(; lSel<(inRoulette.size()-1); ++lSel) {
    if(lVal <= inRoulette[lSel].first) break;
  }
  return inRoulette[lSel].second;
}


/*!
 *  \brief Read a primitive set from a XML node.
 *  \param inNode XML node to read the primitive set from.
 *  \throw IOException If size atribute not present or if the size mismatch.
 */
void GP::WeightedPrimitiveSet::read(XMLNode::Handle& inNode)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "WeightedPrimitiveSet"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <WeightedPrimitiveSet> expected!");
  for(XMLNode::Handle lChild=inNode->getFirstChild();
      lChild != NULL; lChild=lChild->getNextSibling()) {
    if(lChild->getType() == XMLNode::eTag) {
      Primitive::Handle lPrimitive = getPrimitiveByName(lChild->getTagName());
      if(lPrimitive == NULL) {
        std::ostringstream lOSS;
        lOSS << "primitive \"" << lChild->getTagName();
        lOSS << "\" is not in the primitive set!";
        throw Beagle_IOExceptionNodeM(*lChild, lOSS.str());
      }
      lPrimitive->read(lChild);
    }
  }
}


/*!
 *  \brief Write a primitive set into a XML streamer.
 *  \param ioStreamer XML streamer to use to read the primitive set into.
 */
void GP::WeightedPrimitiveSet::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("WeightedPrimitiveSet");
  for(size_type i=0; i<size(); i++) (*this)[i]->write(ioStreamer);
  ioStreamer.closeTag();
}

void GP::WeightedPrimitiveSet::ChangeWeight(unsigned int index, double value)
{
	mWeights[index].first = value;
	Roulette& lAll = mMap[ID_ALL];
	computeRoulette(lAll, ID_ALL);
	Roulette& lBranches = mMap[ID_BRANCHES];
	computeRoulette(lAll, ID_BRANCHES);
	Roulette& lTerm = mMap[0];
	computeRoulette(lTerm, 0);
}
